(* equational systems, minimal -- for use in Biegelinie
   author: Walther Neuper
   050826,
   (c) due to copyright terms
*)

theory EqSystem imports Integrate Rational Root begin

consts

  occur_exactly_in :: 
   "[real list, real list, 'a] => bool" ("_ from _ occur'_exactly'_in _")

  (*descriptions in the related problems*)
  solveForVars       :: "real list => toreall"
  solution           :: "bool list => toreall"

  (*the CAS-command, eg. "solveSystem [x+y=1,y=2] [x,y]"*)
  solveSystem        :: "[bool list, real list] => bool list"

lemma commute_0_equality : \<open>(0 = a) = (a = 0)\<close>
  by auto

axiomatization where
  (*WN0510 see simliar rules 'isolate_' 'separate_' (by RL)
    [bdv_1,bdv_2,bdv_3,bdv_4] work also for 2 and 3 bdvs, ugly !*)
  (* these require lemmas about occur_exactly_in* )
  lemma xxx : 
    \<open>[| [] from [bdv_1, bdv_2, bdv_3, bdv_4] occur_exactly_in a |] ==> (a + b = c) = (b = c + -1 * a)\<close>
    by auto
  ( **)
  separate_bdvs_add:   
    "[| [] from [bdv_1,bdv_2,bdv_3,bdv_4] occur_exactly_in a |] 
		      			     ==> (a + b = c) = (b = c + -1*a)" and
  separate_bdvs0:
    "[| some_of [bdv_1,bdv_2,bdv_3,bdv_4] occur_in b; Not (b=!=0)  |] 
		      			     ==> (a = b) = (a + -1*b = 0)" and
  separate_bdvs_add1:  
    "[| some_of [bdv_1,bdv_2,bdv_3,bdv_4] occur_in c |] 
		      			     ==> (a = b + c) = (a + -1*c = b)" and
  separate_bdvs_add2:
    "[| Not (some_of [bdv_1,bdv_2,bdv_3,bdv_4] occur_in a) |] 
		      			     ==> (a + b = c) = (b = -1*a + c)" and
  separate_bdvs_mult:  
    "[| [] from [bdv_1,bdv_2,bdv_3,bdv_4] occur_exactly_in a; Not (a=!=0) |] 
		      			     ==>(a * b = c) = (b = c / a)" 
axiomatization where (*..if replaced by "and" we get an error in 
  ---  rewrite in [EqSystem,normalise,2x2] --- step "--- 3---";*)
  order_system_NxN:     "[a,b] = [b,a]"
  (*requires rew_ord for termination, eg. ord_simplify_Integral;
    works for lists of any length, interestingly !?!
    CONTRADICTS PROPERTIES OF LIST: take set instead*)

ML \<open>
(** eval functions **)

(*certain variables of a given list occur _all_ in a term
  args: all: ..variables, which are under consideration (eg. the bound vars)
        vs:  variables which must be in t, 
             and none of the others in all must be in t
        t: the term under consideration
 *)
fun occur_exactly_in vs all t =
    let fun occurs_in' a b = Prog_Expr.occurs_in b a
    in foldl and_ (true, map (occurs_in' t) vs)
       andalso not (foldl or_ (false, map (occurs_in' t) 
                                          (subtract op = vs all)))
    end;

(*("occur_exactly_in", ("EqSystem.occur_exactly_in", 
                        eval_occur_exactly_in "#eval_occur_exactly_in_") )*)
fun eval_occur_exactly_in _ "EqSystem.occur_exactly_in"
			  (p as (Const (\<^const_name>\<open>EqSystem.occur_exactly_in\<close>,_) 
				       $ vs $ all $ t)) _ =
    if occur_exactly_in (TermC.isalist2list vs) (TermC.isalist2list all) t
    then SOME ((UnparseC.term @{context} p) ^ " = True",
	       HOLogic.Trueprop $ (TermC.mk_equality (p, @{term True})))
    else SOME ((UnparseC.term @{context} p) ^ " = False",
	       HOLogic.Trueprop $ (TermC.mk_equality (p, @{term False})))
  | eval_occur_exactly_in _ _ _ _ = NONE;
\<close>
calculation occur_exactly_in = \<open>eval_occur_exactly_in "#eval_occur_exactly_in_"\<close>

ML \<open>
(** rewrite order 'ord_simplify_System' **)

(* order wrt. several linear (i.e. without exponents) variables "c", "c_2",..
   which leaves the monomials containing c, c_2,... at the end of an Integral
   and puts the c, c_2,... rightmost within a monomial.

   WN050906 this is a quick and dirty adaption of ord_make_polynomial_in,
   which was most adequate, because it uses size_of_term*)
(**)
local (*. for simplify_System .*)
(**)
open Term;  (* for type order = EQUAL | LESS | GREATER *)

fun pr_ord EQUAL = "EQUAL"
  | pr_ord LESS  = "LESS"
  | pr_ord GREATER = "GREATER";

fun dest_hd' (Const (a, T)) = (((a, 0), T), 0)
  | dest_hd' (Free (ccc, T)) =
    (case Symbol.explode ccc of
	"c"::[] => ((("|||||||||||||||||||||", 0), T), 1)(*greatest string WN*)
      | "c"::"_"::_ => ((("|||||||||||||||||||||", 0), T), 1)
      | _ => (((ccc, 0), T), 1))
  | dest_hd' (Var v) = (v, 2)
  | dest_hd' (Bound i) = ((("", i), dummyT), 3)
  | dest_hd' (Abs (_, T, _)) = ((("", 0), T), 4)
  | dest_hd' _ = raise ERROR "dest_hd': uncovered case in fun.def.";

fun size_of_term' (Free (ccc, _)) =
    (case Symbol.explode ccc of (*WN0510 hack for the bound variables*)
	"c"::[] => 1000
      | "c"::"_"::is => 1000 * ((TermC.int_of_str o implode) is)
      | _ => 1)
  | size_of_term' (Abs (_,_,body)) = 1 + size_of_term' body
  | size_of_term' (f$t) = size_of_term' f  +  size_of_term' t
  | size_of_term' _ = 1;

fun term_ord' pr ctxt (Abs (_, T, t), Abs(_, U, u)) =       (* ~ term.ML *)
    (case term_ord' pr ctxt (t, u) of EQUAL => Term_Ord.typ_ord (T, U) | ord => ord)
  | term_ord' pr ctxt (t, u) =
    (if pr
     then 
       let
         val (f, ts) = strip_comb t and (g, us) = strip_comb u;
         val _ = tracing ("t= f @ ts= \"" ^ UnparseC.term ctxt f ^ "\" @ \"[" ^
           commas (map (UnparseC.term ctxt) ts) ^ "]\"");
         val _ = tracing ("u= g @ us= \"" ^ UnparseC.term ctxt g ^ "\" @ \"[" ^
           commas (map (UnparseC.term ctxt) us) ^ "]\"");
         val _ = tracing ("size_of_term (t, u) = (" ^ string_of_int (size_of_term' t) ^ ", " ^
           string_of_int (size_of_term' u) ^ ")");
         val _ = tracing ("hd_ord (f, g)      = " ^ ((pr_ord o hd_ord) (f, g)) );
    (** )val _ = @{print tracing}{a = "hd_ord (f, g)      = ", z = hd_ord (f, g)}( **)
         val _ = tracing ("terms_ord (ts, us) = " ^ (pr_ord o terms_ord str false ctxt) (ts,us) );
         val _= tracing ("-------");
       in () end
     else ();
    case int_ord (size_of_term' t, size_of_term' u) of
      EQUAL =>
        let val (f, ts) = strip_comb t and (g, us) = strip_comb u 
        in (case hd_ord (f, g) of 
              EQUAL => (terms_ord str pr ctxt) (ts, us) 
            | ord => ord)
        end
	  | ord => ord)
and hd_ord (f, g) =                                        (* ~ term.ML *)
  prod_ord (prod_ord Term_Ord.indexname_ord Term_Ord.typ_ord) int_ord (dest_hd' f, dest_hd' g)
and terms_ord _ pr ctxt (ts, us) = list_ord (term_ord' pr ctxt) (ts, us);
(**)
in
(**)
(*WN0510 for preliminary use in eval_order_system, see case-study mat-eng.tex
fun ord_simplify_System_rev (pr:bool) thy subst tu = 
    (term_ord' pr thy (Library.swap tu) = LESS);*)

(*for the rls's*)
fun ord_simplify_System (pr:bool) thy _(*subst*) (ts, us) = 
    (term_ord' pr thy (TermC.numerals_to_Free ts, TermC.numerals_to_Free us) = LESS);

(**)end;(**)
\<close> ML \<open>
\<close>
setup \<open>Know_Store.add_rew_ords [
  ("ord_simplify_System", ord_simplify_System false)]\<close>

ML \<open>
(** rulesets **)

(*.adapted from 'order_add_mult_in' by just replacing the rew_ord.*)
val order_add_mult_System = 
  Rule_Def.Repeat{
    id = "order_add_mult_System", preconds = [], 
    rew_ord = ("ord_simplify_System", ord_simplify_System false),
    asm_rls = Rule_Set.empty,prog_rls = Rule_Set.Empty, calc = [], errpatts = [],
    rules = [
      \<^rule_thm>\<open>mult.commute\<close>, (* z * w = w * z *)
      \<^rule_thm>\<open>real_mult_left_commute\<close>, (*z1.0 * (z2.0 * z3.0) = z2.0 * (z1.0 * z3.0)*)
      \<^rule_thm>\<open>mult.assoc\<close>,	 (*z1.0 * z2.0 * z3.0 = z1.0 * (z2.0 * z3.0)*)
      \<^rule_thm>\<open>add.commute\<close>, (*z + w = w + z*)
      \<^rule_thm>\<open>add.left_commute\<close>, (*x + (y + z) = y + (x + z)*)
      \<^rule_thm>\<open>add.assoc\<close>	],  (*z1.0 + z2.0 + z3.0 = z1.0 + (z2.0 + z3.0)*)
    program = Rule.Empty_Prog};
\<close>
ML \<open>
(*.adapted from 'norm_Rational' by
  #1 using 'ord_simplify_System' in 'order_add_mult_System'
  #2 NOT using common_nominator_p                          .*)
val norm_System_noadd_fractions = 
  Rule_Def.Repeat {id = "norm_System_noadd_fractions", preconds = [], 
    rew_ord = ("dummy_ord", Rewrite_Ord.function_empty), 
    asm_rls = norm_rat_erls, prog_rls = Rule_Set.Empty, calc = [], errpatts = [],
    rules = [(*sequence given by operator precedence*)
  		Rule.Rls_ discard_minus,
  		Rule.Rls_ powers,
  		Rule.Rls_ rat_mult_divide,
  		Rule.Rls_ expand,
  		Rule.Rls_ reduce_0_1_2,
  		Rule.Rls_ (*order_add_mult #1*) order_add_mult_System,
  		Rule.Rls_ collect_numerals,
  		(*Rule.Rls_ add_fractions_p, #2*)
  		Rule.Rls_ cancel_p],
    program = Rule.Empty_Prog};
\<close>
ML \<open>
(*.adapted from 'norm_Rational' by
  *1* using 'ord_simplify_System' in 'order_add_mult_System'.*)
val norm_System = 
  Rule_Def.Repeat {id = "norm_System", preconds = [], 
    rew_ord = ("dummy_ord", Rewrite_Ord.function_empty), 
    asm_rls = norm_rat_erls, prog_rls = Rule_Set.Empty, calc = [], errpatts = [],
    rules = [(*sequence given by operator precedence*)
  		Rule.Rls_ discard_minus,
  		Rule.Rls_ powers,
  		Rule.Rls_ rat_mult_divide,
  		Rule.Rls_ expand,
  		Rule.Rls_ reduce_0_1_2,
  		Rule.Rls_ (*order_add_mult *1*) order_add_mult_System,
  		Rule.Rls_ collect_numerals,
  		Rule.Rls_ add_fractions_p,
  		Rule.Rls_ cancel_p],
    program = Rule.Empty_Prog};
\<close>
ML \<open>
(*.simplify an equational system BEFORE solving it such that parentheses are
   ( ((u0*v0)*w0) + ( ((u1*v1)*w1) * c + ... +((u4*v4)*w4) * c_4 ) )
ATTENTION: works ONLY for bound variables c, c_1, c_2, c_3, c_4 :ATTENTION
   This is a copy from 'make_ratpoly_in' with respective reductions:
   *0* expand the term, ie. distribute * and / over +
   *1* ord_simplify_System instead of termlessI
   *2* no add_fractions_p (= common_nominator_p_rls !)
   *3* discard_parentheses only for (.*(.*.))
   analoguous to simplify_Integral                                       .*)
val simplify_System_parenthesized = 
  Rule_Set.Sequence {id = "simplify_System_parenthesized", preconds = []:term list, 
    rew_ord = ("dummy_ord", Rewrite_Ord.function_empty),
    asm_rls = Atools_erls, prog_rls = Rule_Set.Empty, calc = [], errpatts = [],
    rules = [
       \<^rule_thm>\<open>distrib_right\<close>, (*"(?z1.0 + ?z2.0) * ?w = ?z1.0 * ?w + ?z2.0 * ?w"*)
	     \<^rule_thm>\<open>add_divide_distrib\<close>, (*"(?x + ?y) / ?z = ?x / ?z + ?y / ?z"*)
	     Rule.Rls_ norm_Rational_noadd_fractions,
	     Rule.Rls_ (*order_add_mult_in*) norm_System_noadd_fractions,
	     \<^rule_thm_sym>\<open>mult.assoc\<close>,
	     Rule.Rls_ collect_bdv, (*from make_polynomial_in WN051031 welldone?*)
	     Rule.Rls_ separate_bdv2,
	     \<^rule_eval>\<open>divide\<close> (Prog_Expr.eval_cancel "#divide_e")],
    program = Rule.Empty_Prog};      
\<close>
ML \<open>
(*.simplify an equational system AFTER solving it;
   This is a copy of 'make_ratpoly_in' with the differences
   *1* ord_simplify_System instead of termlessI           .*)
(*TODO.WN051031 ^^^^^^^^^^ should be in EACH rls contained *)
val simplify_System = 
  Rule_Set.Sequence {id = "simplify_System", preconds = []:term list, 
    rew_ord = ("dummy_ord", Rewrite_Ord.function_empty),
    asm_rls = Atools_erls, prog_rls = Rule_Set.Empty, calc = [], errpatts = [],
    rules = [
      Rule.Rls_ norm_Rational,
	    Rule.Rls_ (*order_add_mult_in*) norm_System (**1**),
	    Rule.Rls_ discard_parentheses,
	    Rule.Rls_ collect_bdv, (*from make_polynomial_in WN051031 welldone?*)
	    Rule.Rls_ separate_bdv2,
	    \<^rule_eval>\<open>divide\<close> (Prog_Expr.eval_cancel "#divide_e")],
    program = Rule.Empty_Prog};      
(*
val simplify_System = 
    Rule_Set.append_rules "simplify_System" simplify_System_parenthesized
	       [\<^rule_thm_sym>\<open>add.assoc\<close>];
*)
\<close>
ML \<open>
val isolate_bdvs = 
  Rule_Def.Repeat {
    id="isolate_bdvs", preconds = [], rew_ord = ("Rewrite_Ord.id_empty", Rewrite_Ord.function_empty), 
    asm_rls = Rule_Set.append_rules "erls_isolate_bdvs" Rule_Set.empty [
      (\<^rule_eval>\<open>occur_exactly_in\<close> (eval_occur_exactly_in "#eval_occur_exactly_in_"))], 
    prog_rls = Rule_Set.Empty, calc = [], errpatts = [],
    rules = [
      \<^rule_thm>\<open>commute_0_equality\<close>,
      \<^rule_thm>\<open>separate_bdvs_add\<close>,
      \<^rule_thm>\<open>separate_bdvs_mult\<close>],
    program = Rule.Empty_Prog};
\<close>
ML \<open>
val isolate_bdvs_4x4 = 
  Rule_Def.Repeat {id="isolate_bdvs_4x4", preconds = [], 
    rew_ord = ("Rewrite_Ord.id_empty", Rewrite_Ord.function_empty), 
    asm_rls = Rule_Set.append_rules "erls_isolate_bdvs_4x4" Rule_Set.empty [
      \<^rule_eval>\<open>occur_exactly_in\<close> (eval_occur_exactly_in "#eval_occur_exactly_in_"),
      \<^rule_eval>\<open>Prog_Expr.ident\<close> (Prog_Expr.eval_ident "#ident_"),
      \<^rule_eval>\<open>Prog_Expr.some_occur_in\<close> (Prog_Expr.eval_some_occur_in "#some_occur_in_"),
      \<^rule_thm>\<open>not_true\<close>,
      \<^rule_thm>\<open>not_false\<close>], 
    prog_rls = Rule_Set.Empty, calc = [], errpatts = [],
    rules = [
      \<^rule_thm>\<open>commute_0_equality\<close>,
      \<^rule_thm>\<open>separate_bdvs0\<close>,
      \<^rule_thm>\<open>separate_bdvs_add1\<close>,
      \<^rule_thm>\<open>separate_bdvs_add2\<close>,
      \<^rule_thm>\<open>separate_bdvs_mult\<close>],
    program = Rule.Empty_Prog};

\<close>
ML \<open>

(*.order the equations in a system such, that a triangular system (if any)
   appears as [..c_4 = .., ..., ..., ..c_1 + ..c_2 + ..c_3 ..c_4 = ..].*)
val order_system = 
  Rule_Def.Repeat {id="order_system", preconds = [], 
	  rew_ord = ("ord_simplify_System", ord_simplify_System false), 
	  asm_rls = Rule_Set.Empty, prog_rls = Rule_Set.Empty, calc = [], errpatts = [],
	  rules = [
      \<^rule_thm>\<open>order_system_NxN\<close>],
	  program = Rule.Empty_Prog};

val prls_triangular = 
  Rule_Def.Repeat {
    id="prls_triangular", preconds = [], rew_ord = ("Rewrite_Ord.id_empty", Rewrite_Ord.function_empty), 
    asm_rls = Rule_Def.Repeat {
      id="erls_prls_triangular", preconds = [], rew_ord = ("Rewrite_Ord.id_empty", Rewrite_Ord.function_empty), 
      asm_rls = Rule_Set.Empty, prog_rls = Rule_Set.Empty, calc = [], errpatts = [],
      rules = [(*for precond NTH_CONS ...*)
         \<^rule_eval>\<open>less\<close> (Prog_Expr.eval_equ "#less_"),
         \<^rule_eval>\<open>plus\<close> (Calc_Binop.numeric "#add_"),
         \<^rule_eval>\<open>occur_exactly_in\<close> (eval_occur_exactly_in "#eval_occur_exactly_in_")],
         (*immediately repeated rewrite pushes '+' into precondition !*)
      program = Rule.Empty_Prog}, 
    prog_rls = Rule_Set.Empty, calc = [], errpatts = [],
    rules = [
      \<^rule_thm>\<open>NTH_CONS\<close>,
      \<^rule_eval>\<open>plus\<close> (Calc_Binop.numeric "#add_"),
      \<^rule_thm>\<open>NTH_NIL\<close>,
      \<^rule_thm>\<open>tl_Cons\<close>,
      \<^rule_thm>\<open>tl_Nil\<close>,
      \<^rule_eval>\<open>occur_exactly_in\<close> (eval_occur_exactly_in "#eval_occur_exactly_in_")],
    program = Rule.Empty_Prog};
\<close>
ML \<open>

(*WN060914 quickly created for 4x4; 
 more similarity to prls_triangular desirable*)
val prls_triangular4 = 
  Rule_Def.Repeat {
  id="prls_triangular4", preconds = [], rew_ord = ("Rewrite_Ord.id_empty", Rewrite_Ord.function_empty), 
  asm_rls = Rule_Def.Repeat {
    id="erls_prls_triangular4", preconds = [], rew_ord = ("Rewrite_Ord.id_empty", Rewrite_Ord.function_empty), 
    asm_rls = Rule_Set.Empty, prog_rls = Rule_Set.Empty, calc = [], errpatts = [],
    rules = [(*for precond NTH_CONS ...*)
  	   \<^rule_eval>\<open>less\<close> (Prog_Expr.eval_equ "#less_"),
  	   \<^rule_eval>\<open>plus\<close> (**)(Calc_Binop.numeric "#add_")],
  	   (*immediately repeated rewrite pushes '+' into precondition !*)
    program = Rule.Empty_Prog}, 
  prog_rls = Rule_Set.Empty, calc = [], errpatts = [],
  rules = [
    \<^rule_thm>\<open>NTH_CONS\<close>,
    \<^rule_eval>\<open>plus\<close> (**)(Calc_Binop.numeric "#add_"),
    \<^rule_thm>\<open>NTH_NIL\<close>,
    \<^rule_thm>\<open>tl_Cons\<close>,
    \<^rule_thm>\<open>tl_Nil\<close>,
    \<^rule_eval>\<open>occur_exactly_in\<close> (eval_occur_exactly_in "#eval_occur_exactly_in_")],
  program = Rule.Empty_Prog};
\<close>

rule_set_knowledge
  simplify_System_parenthesized = \<open>prep_rls' simplify_System_parenthesized\<close> and
  simplify_System = \<open>prep_rls' simplify_System\<close> and
  isolate_bdvs = \<open>prep_rls' isolate_bdvs\<close> and
  isolate_bdvs_4x4 = \<open>prep_rls' isolate_bdvs_4x4\<close> and 
  order_system = \<open>prep_rls' order_system\<close> and 
  order_add_mult_System = \<open>prep_rls' order_add_mult_System\<close> and
  norm_System_noadd_fractions = \<open>prep_rls' norm_System_noadd_fractions\<close> and
  norm_System = \<open>prep_rls' norm_System\<close>


section \<open>Problems\<close>

problem pbl_equsys : "system" =
  \<open>Rule_Set.append_rules "empty" Rule_Set.empty [(*for preds in where_*)]\<close>
  CAS: "solveSystem e_s v_s"
  Given: "equalities e_s" "solveForVars v_s"
  Find: "solution ss'''" (*''' is copy-named*)

problem pbl_equsys_lin : "LINEAR/system" =
  \<open>Rule_Set.append_rules "empty" Rule_Set.empty [(*for preds in where_*)]\<close>
  CAS: "solveSystem e_s v_s"
  Given: "equalities e_s" "solveForVars v_s"
  (*TODO.WN050929 check linearity*)
  Find: "solution ss'''"

problem pbl_equsys_lin_2x2: "2x2/LINEAR/system" =
  \<open>Rule_Set.append_rules "prls_2x2_linear_system" Rule_Set.empty 
    [\<^rule_thm>\<open>LENGTH_CONS\<close>,
      \<^rule_thm>\<open>LENGTH_NIL\<close>,
      \<^rule_eval>\<open>plus\<close> (**)(Calc_Binop.numeric "#add_"),
      \<^rule_eval>\<open>HOL.eq\<close> (Prog_Expr.eval_equal "#equal_")]\<close>
  CAS: "solveSystem e_s v_s"
  Given: "equalities e_s" "solveForVars v_s"
  Where: "Length (e_s:: bool list) = 2" "Length v_s = 2"
  Find: "solution ss'''"

problem pbl_equsys_lin_2x2_tri : "triangular/2x2/LINEAR/system" =
  \<open>prls_triangular\<close>
  Method_Ref: "EqSystem/top_down_substitution/2x2"
  CAS: "solveSystem e_s v_s"
  Given: "equalities e_s" "solveForVars v_s"
  Where:
    "(tl v_s) from v_s occur_exactly_in (NTH 1 (e_s::bool list))"
    "    v_s  from v_s occur_exactly_in (NTH 2 (e_s::bool list))"
  Find: "solution ss'''"

problem pbl_equsys_lin_2x2_norm : "normalise/2x2/LINEAR/system" =
  \<open>Rule_Set.append_rules "empty" Rule_Set.empty [(*for preds in where_*)]\<close>
  Method_Ref: "EqSystem/normalise/2x2"
  CAS: "solveSystem e_s v_s"
  Given: "equalities e_s" "solveForVars v_s"
  Find: "solution ss'''"

problem pbl_equsys_lin_3x3 : "3x3/LINEAR/system" =
  \<open>Rule_Set.append_rules "prls_3x3_linear_system" Rule_Set.empty 
    [\<^rule_thm>\<open>LENGTH_CONS\<close>,
      \<^rule_thm>\<open>LENGTH_NIL\<close>,
      \<^rule_eval>\<open>plus\<close> (**)(Calc_Binop.numeric "#add_"),
      \<^rule_eval>\<open>HOL.eq\<close> (Prog_Expr.eval_equal "#equal_")]\<close>
  CAS: "solveSystem e_s v_s"
  Given: "equalities e_s" "solveForVars v_s"
  Where: "Length (e_s:: bool list) = 3" "Length v_s = 3"
  Find: "solution ss'''"

problem pbl_equsys_lin_4x4 : "4x4/LINEAR/system" =
  \<open>Rule_Set.append_rules "prls_4x4_linear_system" Rule_Set.empty 
    [\<^rule_thm>\<open>LENGTH_CONS\<close>,
      \<^rule_thm>\<open>LENGTH_NIL\<close>,
      \<^rule_eval>\<open>plus\<close> (**)(Calc_Binop.numeric "#add_"),
      \<^rule_eval>\<open>HOL.eq\<close> (Prog_Expr.eval_equal "#equal_")]\<close>
  CAS: "solveSystem e_s v_s"
  Given: "equalities e_s" "solveForVars v_s"
  Where: "Length (e_s:: bool list) = 4" "Length v_s = 4"
  Find: "solution ss'''"

problem pbl_equsys_lin_4x4_tri : "triangular/4x4/LINEAR/system" =
  \<open>Rule_Set.append_rules "prls_tri_4x4_lin_sys" prls_triangular
    [\<^rule_eval>\<open>Prog_Expr.occurs_in\<close> (Prog_Expr.eval_occurs_in "")]\<close>
  Method_Ref: "EqSystem/top_down_substitution/4x4"
  CAS: "solveSystem e_s v_s"
  Given: "equalities e_s" "solveForVars v_s"
  Where: (*accepts missing variables up to diagional form*)
    "(NTH 1 (v_s::real list)) occurs_in (NTH 1 (e_s::bool list))"
    "(NTH 2 (v_s::real list)) occurs_in (NTH 2 (e_s::bool list))"
    "(NTH 3 (v_s::real list)) occurs_in (NTH 3 (e_s::bool list))"
    "(NTH 4 (v_s::real list)) occurs_in (NTH 4 (e_s::bool list))"
  Find: "solution ss'''"

problem pbl_equsys_lin_4x4_norm : "normalise/4x4/LINEAR/system" =
  \<open>Rule_Set.append_rules "empty" Rule_Set.empty [(*for preds in where_*)]\<close>
  Method_Ref: "EqSystem/normalise/4x4"
  CAS: "solveSystem e_s v_s"
  Given: "equalities e_s" "solveForVars v_s"
  (*Length is checked 1 level above*)
  Find: "solution ss'''"

ML \<open>
(*this is for NTH only*)
val prog_rls = Rule_Def.Repeat {id="srls_normalise_4x4", 
		preconds = [], 
		rew_ord = ("termlessI",termlessI), 
		asm_rls = Rule_Set.append_rules "erls_in_srls_IntegrierenUnd.." Rule_Set.empty
				  [(*for asm in NTH_CONS ...*)
				   \<^rule_eval>\<open>less\<close> (Prog_Expr.eval_equ "#less_"),
				   (*2nd NTH_CONS pushes n+-1 into asms*)
				   \<^rule_eval>\<open>plus\<close> (**)(Calc_Binop.numeric "#add_")
				   ], 
		prog_rls = Rule_Set.Empty, calc = [], errpatts = [],
		rules = [\<^rule_thm>\<open>NTH_CONS\<close>,
			 \<^rule_eval>\<open>plus\<close> (**)(Calc_Binop.numeric "#add_"),
			 \<^rule_thm>\<open>NTH_NIL\<close>],
		program = Rule.Empty_Prog};
\<close>

section \<open>Methods\<close>

method met_eqsys : "EqSystem" =
  \<open>{rew_ord="tless_true", rls' = Rule_Set.Empty, calc = [], prog_rls = Rule_Set.Empty, where_rls = Rule_Set.Empty,
    errpats = [], rew_rls = Rule_Set.Empty}\<close>

method met_eqsys_topdown : "EqSystem/top_down_substitution" =
  \<open>{rew_ord="tless_true", rls' = Rule_Set.Empty, calc = [], prog_rls = Rule_Set.Empty, where_rls = Rule_Set.Empty,
    errpats = [], rew_rls = Rule_Set.Empty}\<close>

partial_function (tailrec) solve_system :: "bool list => real list => bool list"
  where
"solve_system e_s v_s = (
  let
    e_1 = Take (hd e_s);                                                         
    e_1 = (
      (Try (Rewrite_Set_Inst [(''bdv_1'', hd v_s), (''bdv_2'', hd (tl v_s))] ''isolate_bdvs'')) #>                   
      (Try (Rewrite_Set_Inst [(''bdv_1'', hd v_s), (''bdv_2'', hd (tl v_s))] ''simplify_System''))
      ) e_1;                 
    e_2 = Take (hd (tl e_s));                                                    
    e_2 = (
      (Substitute [e_1]) #>                                                 
      (Try (Rewrite_Set_Inst [(''bdv_1'', hd v_s), (''bdv_2'', hd (tl v_s))] ''simplify_System_parenthesized'' )) #>      
      (Try (Rewrite_Set_Inst [(''bdv_1'', hd v_s), (''bdv_2'', hd (tl v_s))] ''isolate_bdvs'' )) #>                   
      (Try (Rewrite_Set_Inst [(''bdv_1'', hd v_s), (''bdv_2'', hd (tl v_s))] ''simplify_System'' ))
      ) e_2;                 
    e__s = Take [e_1, e_2]                                                       
  in
    Try (Rewrite_Set ''order_system'' ) e__s)                              "

method met_eqsys_topdown_2x2 : "EqSystem/top_down_substitution/2x2" =
  \<open>{rew_ord="ord_simplify_System", rls' = Rule_Set.Empty, calc = [], 
    prog_rls = Rule_Set.append_rules "srls_top_down_2x2" Rule_Set.empty
        [\<^rule_thm>\<open>hd_thm\<close>,
          \<^rule_thm>\<open>tl_Cons\<close>,
          \<^rule_thm>\<open>tl_Nil\<close>], 
    where_rls = prls_triangular, errpats = [], rew_rls = Rule_Set.Empty}\<close>
  Program: solve_system.simps
  Given: "equalities e_s" "solveForVars v_s"
  Where:
    "(tl v_s) from v_s occur_exactly_in (NTH 1 (e_s::bool list))"
    "    v_s  from v_s occur_exactly_in (NTH 2 (e_s::bool list))"
  Find: "solution ss'''"

method met_eqsys_norm : "EqSystem/normalise" =
  \<open>{rew_ord="tless_true", rls' = Rule_Set.Empty, calc = [], prog_rls = Rule_Set.Empty, where_rls = Rule_Set.Empty,
    errpats = [], rew_rls = Rule_Set.Empty}\<close>

partial_function (tailrec) solve_system2 :: "bool list => real list => bool list"
  where
"solve_system2 e_s v_s = (
  let
    e__s = (
      (Try (Rewrite_Set ''norm_Rational'' )) #>
      (Try (Rewrite_Set_Inst [(''bdv_1'', hd v_s), (''bdv_2'', hd (tl v_s))] ''simplify_System_parenthesized'' )) #>
      (Try (Rewrite_Set_Inst [(''bdv_1'', hd v_s), (''bdv_2'', hd (tl v_s))] ''isolate_bdvs'' )) #>
      (Try (Rewrite_Set_Inst [(''bdv_1'', hd v_s), (''bdv_2'', hd (tl v_s))] ''simplify_System_parenthesized'' )) #>
      (Try (Rewrite_Set ''order_system'' ))
      ) e_s
  in
    SubProblem (''EqSystem'', [''LINEAR'', ''system''], [''no_met''])
      [BOOL_LIST e__s, REAL_LIST v_s])"

method met_eqsys_norm_2x2 : "EqSystem/normalise/2x2" =
  \<open>{rew_ord="tless_true", rls' = Rule_Set.Empty, calc = [], 
    prog_rls = Rule_Set.append_rules "srls_normalise_2x2" Rule_Set.empty
        [\<^rule_thm>\<open>hd_thm\<close>,
          \<^rule_thm>\<open>tl_Cons\<close>,
          \<^rule_thm>\<open>tl_Nil\<close>], 
    where_rls = Rule_Set.Empty, errpats = [], rew_rls = Rule_Set.Empty}\<close>
  Program: solve_system2.simps
  Given: "equalities e_s" "solveForVars v_s"
  Find: "solution ss'''"

partial_function (tailrec) solve_system3 :: "bool list => real list => bool list"
  where
"solve_system3 e_s v_s = (
  let
    e__s = (
      (Try (Rewrite_Set ''norm_Rational'' )) #>
      (Repeat (Rewrite ''commute_0_equality'' )) #>
      (Try (Rewrite_Set_Inst [(''bdv_1'', NTH 1 v_s), (''bdv_2'', NTH 2 v_s ),
        (''bdv_3'', NTH 3 v_s), (''bdv_3'', NTH 4 v_s )] ''simplify_System_parenthesized'' )) #>
      (Try (Rewrite_Set_Inst [(''bdv_1'', NTH 1 v_s), (''bdv_2'', NTH 2 v_s ),
        (''bdv_3'', NTH 3 v_s), (''bdv_3'', NTH 4 v_s )] ''isolate_bdvs_4x4'' )) #>
      (Try (Rewrite_Set_Inst [(''bdv_1'', NTH 1 v_s), (''bdv_2'', NTH 2 v_s ),
        (''bdv_3'', NTH 3 v_s), (''bdv_3'', NTH 4 v_s )] ''simplify_System_parenthesized'' )) #>
      (Try (Rewrite_Set ''order_system''))
      )  e_s
  in
    SubProblem (''EqSystem'', [''LINEAR'', ''system''], [''no_met''])
      [BOOL_LIST e__s, REAL_LIST v_s])"

method met_eqsys_norm_4x4 : "EqSystem/normalise/4x4" =
  \<open>{rew_ord="tless_true", rls' = Rule_Set.Empty, calc = [],
    prog_rls =
      Rule_Set.append_rules "srls_normalise_4x4" prog_rls
        [\<^rule_thm>\<open>hd_thm\<close>, \<^rule_thm>\<open>tl_Cons\<close>, \<^rule_thm>\<open>tl_Nil\<close>],
    where_rls = Rule_Set.Empty, errpats = [], rew_rls = Rule_Set.Empty}\<close>
  Program: solve_system3.simps
  Given: "equalities e_s" "solveForVars v_s"
  Find: "solution ss'''"
  (*STOPPED.WN06? met ["EqSystem", "normalise", "4x4"] #>#>#>#>#>#>#>#>#>#>#>#>#>@*)

partial_function (tailrec) solve_system4 :: "bool list => real list => bool list"
  where
"solve_system4 e_s v_s = (
  let
    e_1 = NTH 1 e_s;
    e_2 = Take (NTH 2 e_s);
    e_2 = (
      (Substitute [e_1]) #>
      (Try (Rewrite_Set_Inst [(''bdv_1'',NTH 1 v_s), (''bdv_2'',NTH 2 v_s),
        (''bdv_3'',NTH 3 v_s), (''bdv_4'',NTH 4 v_s)] ''simplify_System_parenthesized'' )) #>
      (Try (Rewrite_Set_Inst [(''bdv_1'',NTH 1 v_s), (''bdv_2'',NTH 2 v_s),
        (''bdv_3'',NTH 3 v_s), (''bdv_4'',NTH 4 v_s)] ''isolate_bdvs'' )) #>
      (Try (Rewrite_Set_Inst [(''bdv_1'',NTH 1 v_s), (''bdv_2'',NTH 2 v_s),
        (''bdv_3'',NTH 3 v_s), (''bdv_4'',NTH 4 v_s)] ''norm_Rational'' ))
      ) e_2
  in
    [e_1, e_2, NTH 3 e_s, NTH 4 e_s])"

method met_eqsys_topdown_4x4 : "EqSystem/top_down_substitution/4x4" =
  \<open>{rew_ord="ord_simplify_System", rls' = Rule_Set.Empty, calc = [], 
    prog_rls = Rule_Set.append_rules "srls_top_down_4x4" prog_rls [], 
    where_rls = Rule_Set.append_rules "prls_tri_4x4_lin_sys" prls_triangular
      [\<^rule_eval>\<open>Prog_Expr.occurs_in\<close> (Prog_Expr.eval_occurs_in "")], 
    errpats = [], rew_rls = Rule_Set.Empty}\<close>
  (*FIXXXXME.WN060916: this script works ONLY for exp 7.79 #>#>#>#>#>#>#>#>#>#>*)
  Program: solve_system4.simps
  Given: "equalities e_s" "solveForVars v_s"
  Where: (*accepts missing variables up to diagonal form*)
    "(NTH 1 (v_s::real list)) occurs_in (NTH 1 (e_s::bool list))"
    "(NTH 2 (v_s::real list)) occurs_in (NTH 2 (e_s::bool list))"
    "(NTH 3 (v_s::real list)) occurs_in (NTH 3 (e_s::bool list))"
    "(NTH 4 (v_s::real list)) occurs_in (NTH 4 (e_s::bool list))"
  Find: "solution ss'''"

ML \<open>
\<close> ML \<open>
\<close> ML \<open>
\<close>
end
